@@ -1,10 +1,8 @@
-#!/usr/bin/env python
-
 ## system-config-printer
 
-## Copyright (C) 2006, 2007 Red Hat, Inc.
+## Copyright (C) 2006, 2007, 2008 Red Hat, Inc.
 ## Copyright (C) 2006 Florian Festi <ffesti@redhat.com>
-## Copyright (C) 2006, 2007 Tim Waugh <twaugh@redhat.com>
+## Copyright (C) 2006, 2007, 2008 Tim Waugh <twaugh@redhat.com>
 
 ## This program is free software; you can redistribute it and/or modify
 ## it under the terms of the GNU General Public License as published by
@@ -22,6 +20,8 @@
 
 import cups, pprint, os, tempfile, re
 from rhpl.translate import _, N_
+import locale
+from debug import *
 
 class Printer:
 
@@ -34,26 +34,7 @@ class Printer:
         self.name = name
         self.connection = connection
         self.class_members = []
-        self.device_uri = kw.get('device-uri', "")
-        self.info = kw.get('printer-info', "")
-        self.is_shared = kw.get('printer-is-shared', None)
-        self.location = kw.get('printer-location', "")
-        self.make_and_model = kw.get('printer-make-and-model', "")
-        self.state = kw.get('printer-state', 0)
-        self.type = kw.get('printer-type', 0)
-        self.uri_supported = kw.get('printer-uri-supported', "")
-        self._expand_flags()
-
-        self.state_description = self.printer_states.get(
-            self.state, _("Unknown"))
-
-        self._getAttributes()
-
-        self.enabled = self.state != cups.IPP_PRINTER_STOPPED
-
-        if self.is_shared is None:
-            self.is_shared = not self.not_shared
-        del self.not_shared
+        self.update (**kw)
 
         if self.is_class:
             self._ppd = False
@@ -75,9 +56,27 @@ class Printer:
                 setattr(self, attr_name,
                         bool(self.type & getattr(cups, name)))
 
-    def _getAttributes(self):
+    def update(self, **kw):
+        self.state = kw.get('printer-state', 0)
+        self.enabled = self.state != cups.IPP_PRINTER_STOPPED
+        self.state_description = self.printer_states.get(
+            self.state, _("Unknown"))
+        self.device_uri = kw.get('device-uri', "")
+        self.info = kw.get('printer-info', "")
+        self.is_shared = kw.get('printer-is-shared', None)
+        self.location = kw.get('printer-location', "")
+        self.make_and_model = kw.get('printer-make-and-model', "")
+        self.type = kw.get('printer-type', 0)
+        self.uri_supported = kw.get('printer-uri-supported', "")
+        self._expand_flags()
+        if self.is_shared is None:
+            self.is_shared = not self.not_shared
+        del self.not_shared
+
+    def getAttributes(self):
         attrs = self.connection.getPrinterAttributes(self.name)
         self.attributes = {}
+        self.other_attributes = {}
         self.possible_attributes = {
             'landscape' : ('False', ['True', 'False']),
             'page-border' : ('none', ['none', 'single', 'single-thick',
@@ -91,18 +90,30 @@ class Printer:
                             "printer-op-policy", # handled below
                             "notify-events", # cannot be set
                             "document-format", # cannot be set
-                            "notify-lease-duration", # cannot be set
-                            "job-hold-until"]: # no sensible way to implement
+                            "notify-lease-duration"]: # cannot be set
                     continue 
 
                 supported = attrs.get(name + "-supported", None) or \
                             self.possible_attributes.get(name, None) or \
                             ""
+
+                # Convert a list into a comma-separated string, since
+                # it can only really have been misinterpreted as a list
+                # by CUPS.
+                if isinstance (value, list):
+                    value = reduce (lambda x, y: x+','+y, value)
+
                 self.attributes[name] = value
                     
                 if attrs.has_key(name+"-supported"):
                     self.possible_attributes[name] = (
                         value, attrs[name+"-supported"]) 
+            elif (not key.endswith ("-supported") and
+                  key != 'job-sheets-default' and
+                  key != 'printer-error-policy' and
+                  key != 'printer-op-policy' and
+                  not key.startswith ('requesting-user-name-')):
+                self.other_attributes[key] = value
         
         self.job_sheet_start, self.job_sheet_end = attrs.get(
             'job-sheets-default', ('none', 'none'))
@@ -122,6 +133,13 @@ class Printer:
         elif attrs.has_key('requesting-user-name-denied'):
             self.except_users = attrs['requesting-user-name-denied']
         self.except_users_string = ', '.join(self.except_users)
+        if (attrs.has_key ('device-uri') and
+            attrs['device-uri'].startswith ("smb:")):
+            # Don't update device-uri for smb printers as we need to have
+            # the original from printers.conf, in case it has authentication
+            # details in it.
+            attrs['device-uri'] = self.device_uri
+        self.update (**attrs)
 
     def getServer(self):
         """return Server URI or None"""
@@ -152,6 +170,11 @@ class Printer:
         return self._ppd
 
     def setOption(self, name, value):
+        if isinstance (value, float):
+            radixchar = locale.nl_langinfo (locale.RADIXCHAR)
+            if radixchar != '.':
+                # Convert floats to strings, being careful with decimal points.
+                value = str (value).replace (radixchar, '.')
         self.connection.addPrinterOptionDefault(self.name, name, value)
 
     def unsetOption(self, name):
@@ -202,8 +225,8 @@ class Printer:
         else:
             self.connection.setPrinterUsersAllowed(self.name, except_users)
 
-    def testsQueued(self):
-        """Returns a list of job IDs for test pages in the queue for this
+    def jobsQueued(self, only_tests=False):
+        """Returns a list of job IDs for jobs in the queue for this
         printer."""
         ret = []
         try:
@@ -220,10 +243,71 @@ class Printer:
             if uri != self.name:
                 continue
 
-            if attrs.has_key ('job-name') and attrs['job-name'] == 'Test Page':
+            if (not only_tests or
+                (attrs.has_key ('job-name') and
+                 attrs['job-name'] == 'Test Page')):
                 ret.append (id)
         return ret
 
+    def testsQueued(self):
+        """Returns a list of job IDs for test pages in the queue for this
+        printer."""
+        return self.jobsQueued (only_tests=True)
+
+    def setAsDefault(self):
+        self.connection.setDefault(self.name)
+
+        # Also need to check system-wide lpoptions because that's how
+        # previous Fedora versions set the default (bug #217395).
+        (tmpfd, tmpfname) = tempfile.mkstemp ()
+        try:
+            resource = "/admin/conf/lpoptions"
+
+            try:
+                # Specifying the fd is allowed in pycups >= 1.9.38
+                self.connection.getFile(resource, fd=tmpfd)
+            except TypeError:
+                self.connection.getFile(resource, tmpfname)
+        except cups.HTTPError, (s,):
+            try:
+                os.remove (tmpfname)
+            except OSError:
+                pass
+
+            if s == cups.HTTP_NOT_FOUND:
+                return
+
+            raise cups.HTTPError (s)
+
+        lines = file (tmpfname).readlines ()
+        changed = False
+        i = 0
+        for line in lines:
+            if line.startswith ("Default "):
+                # This is the system-wide default.
+                name = line.split (' ')[1]
+                if name != self.name:
+                    # Stop it from over-riding the server default.
+                    lines[i] = "Dest " + line[8:]
+                    changed = True
+                i += 1
+
+        if changed:
+            file (tmpfname, 'w').writelines (lines)
+            try:
+                self.connection.putFile (resource, tmpfname)
+            except cups.HTTPError, (s,):
+                os.remove (tmpfname)
+                return
+
+            # Now reconnect because the server needs to reload.
+            self.reconnect ()
+
+        try:
+            os.remove (tmpfname)
+        except OSError:
+            pass
+
 def getPrinters(connection):
     printers = connection.getPrinters()
     classes = connection.getClasses()
@@ -242,6 +326,19 @@ def getPrinters(connection):
                 printers_conf = PrintersConf(connection)
             if printers_conf.device_uris.has_key(name):
                 printer.device_uri = printers_conf.device_uris[name]
+        if not printer.__dict__.has_key ('discovered'):
+            # The CUPS_PRINTER_DISCOVERED flag is new in pycups-1.9.37.
+            printer.discovered = False
+            if printer.device_uri.startswith ("ipp:"):
+                # ipp: Queues can be automatically created as a reaction
+                # to a queue broadcasted by a remote CUPS server. These
+                # queues are not listed in printers,conf. Mark them so
+                # that we can decide which queue entries in the main window
+                # should be editable/deletable and which not
+                if not printers_conf:
+                    printers_conf = PrintersConf(connection)
+                if not printers_conf.device_uris.has_key(name):
+                    printer.discovered = True
     return printers
 
 def parseDeviceID (id):
@@ -251,8 +348,6 @@ def parseDeviceID (id):
         if piece.find(":") == -1:
             continue
         name, value = piece.split(":",1)
-        if name=="CMD":
-            value = value.split(',') 
         id_dict[name] = value
     if id_dict.has_key ("MANUFACTURER"):
         id_dict.setdefault("MFG", id_dict["MANUFACTURER"])
@@ -262,6 +357,7 @@ def parseDeviceID (id):
         id_dict.setdefault("CMD", id_dict["COMMAND SET"])
     for name in ["MFG", "MDL", "CMD", "CLS", "DES", "SN", "S", "P", "J"]:
         id_dict.setdefault(name, "")
+    id_dict["CMD"] = id_dict["CMD"].split(',') 
     return id_dict
 
 class Device:
@@ -291,7 +387,8 @@ class Device:
                 return -1
             return 1
         if not self.is_class and (self.type != other.type):
-            # "hp" before * before "parallel" before "serial"
+            # "hp"/"hpfax" before * before "usb" before "parallel" before
+            # "serial"
             if other.type == "serial":
                 return -1
             if self.type == "serial":
@@ -300,9 +397,13 @@ class Device:
                 return -1
             if self.type == "parallel":
                 return 1
-            if other.type == "hp":
+            if other.type == "usb":
+                return -1
+            if self.type == "usb":
                 return 1
-            if self.type == "hp":
+            if other.type == "hp" or other.type == "hpfax":
+                return 1
+            if self.type == "hp" or self.type == "hpfax":
                 return -1
         result = cmp(bool(self.id), bool(other.id))
         if not result:
@@ -318,16 +419,21 @@ class PrintersConf:
 
     def fetch(self, file):
         fd, filename = tempfile.mkstemp("printer.conf")
-        os.close(fd)
         try:
-            self.connection.getFile(file, filename)
+            try:
+                # Specifying the fd is allowed in pycups >= 1.9.38
+                self.connection.getFile(file, fd=fd)
+            except TypeError:
+                self.connection.getFile(file, filename)
         except cups.HTTPError, e:
+            os.close(fd)
             if (e.args[0] == cups.HTTP_UNAUTHORIZED or
                 e.args[0] == cups.HTTP_NOT_FOUND):
                 return []
             else:
                 raise e
 
+        os.close(fd)
         lines = open(filename).readlines()
         os.unlink(filename)
         return lines
@@ -350,13 +456,7 @@ class PrintersConf:
                 if line.strip().find("</Printer>") != -1:
                     current_printer = None
 
-def match(s1, s2):
-    if s1==s2: return len(s1)
-    for nr, (c1, c2) in enumerate(zip(s1, s2)):
-        if c1!=c2: return nr
-    return min(len(s1), len(s2))
-
-def getDevices(connection, current_uri=None):
+def getDevices(connection):
     """
     raise cups.IPPError
     """
@@ -364,16 +464,31 @@ def getDevices(connection, current_uri=None):
     for uri, data in devices.iteritems():
         device = Device(uri, **data)
         devices[uri] = device
-    if current_uri and not devices.has_key(current_uri):
-        device = Device(current_uri)
-        uri_matches = [(match(uri, current_uri), uri)
-                       for uri in devices.iterkeys()]
-                      # returns list of (match length, uri) 
-        m, uri = max(uri_matches)
-        device.info = devices[uri].info
-        devices[current_uri] = device
+        if device.info != 'Unknown' and device.make_and_model == 'Unknown':
+            device.make_and_model = device.info
     return devices
 
+def activateNewPrinter(connection, name):
+    """Set a new printer enabled, accepting jobs, and
+    (if necessary) the default printer."""
+    connection.enablePrinter (name)
+    connection.acceptJobs (name)
+
+    # Set as the default if there is not already a default printer.
+    default_is_set = False
+    try:
+        if connection.getDefault () != None:
+            default_is_set = True
+    except AttributeError: # getDefault appeared in pycups-1.9.31
+        dests = connection.getDests ()
+        # If a default printer is set it will be available from
+        # key (None,None).
+        if dests.has_key ((None, None)):
+            default_is_set = True
+
+    if not default_is_set:
+        connection.setDefault (name)
+
 def getPPDGroupOptions(group):
     options = group.options[:]
     for g in group.subgroups:
@@ -395,7 +510,7 @@ def copyPPDOptions(ppd1, ppd2):
             for choice in new_option.choices:
                 if choice["choice"]==value:
                     ppd2.markOption(new_option.keyword, value)
-                    print "set %s = %s" % (new_option.keyword, value)
+                    debugprint ("set %s = %s" % (new_option.keyword, value))
                     
 def setPPDPageSize(ppd, language):
     # Just set the page size to A4 or Letter, that's all.
@@ -407,10 +522,158 @@ def setPPDPageSize(ppd, language):
             size = 'Letter'
     try:
         ppd.markOption ('PageSize', size)
-        print "set PageSize = %s" % size
+        debugprint ("set PageSize = %s" % size)
     except:
-        print "Failed to set PageSize " \
-              "(%s not available?)" % size
+        debugprint ("Failed to set PageSize (%s not available?)" % size)
+
+def missingPackagesAndExecutables(ppd):
+    """Check that all relevant executables for a PPD are installed.
+
+    ppd: cups.PPD object"""
+
+    # First, a local function.  How to check that something exists
+    # in a path:
+    def pathcheck (name, path="/usr/bin:/bin"):
+        # Strip out foomatic '%'-style place-holders.
+        p = name.find ('%')
+        if p != -1:
+            name = name[:p]
+        if len (name) == 0:
+            return "true"
+        if name[0] == '/':
+            if os.access (name, os.X_OK):
+                debugprint ("%s: found" % name)
+                return name
+            else:
+                debugprint ("%s: NOT found" % name)
+                return None
+        if name.find ("=") != -1:
+            return "builtin"
+        if name in [ ":", ".", "[", "alias", "bind", "break", "cd",
+                     "continue", "declare", "echo", "else", "eval",
+                     "exec", "exit", "export", "fi", "if", "kill", "let",
+                     "local", "popd", "printf", "pushd", "pwd", "read",
+                     "readonly", "set", "shift", "shopt", "source",
+                     "test", "then", "trap", "type", "ulimit", "umask",
+                     "unalias", "unset", "wait" ]:
+            return "builtin"
+        for component in path.split (':'):
+            file = component.rstrip (os.path.sep) + os.path.sep + name
+            if os.access (file, os.X_OK):
+                debugprint ("%s: found" % file)
+                return file
+        debugprint ("%s: NOT found in %s" % (name,path))
+        return None
+
+    pkgs_to_install = []
+    exes_to_install = []
+
+    # Find a 'FoomaticRIPCommandLine' attribute.
+    exe = exepath = None
+    attr = ppd.findAttr ('FoomaticRIPCommandLine')
+    if attr:
+        # Foomatic RIP command line to check.
+        cmdline = attr.value.replace ('&&\n', '')
+        cmdline = cmdline.replace ('&quot;', '"')
+        cmdline = cmdline.replace ('&lt;', '<')
+        cmdline = cmdline.replace ('&gt;', '>')
+        if (cmdline.find ("(") != -1 or
+            cmdline.find ("&") != -1):
+            # Don't try to handle sub-shells or unreplaced HTML entities.
+            cmdline = ""
+
+        # Strip out foomatic '%'-style place-holders
+        pipes = cmdline.split (';')
+        for pipe in pipes:
+            cmds = pipe.strip ().split ('|')
+            for cmd in cmds:
+                args = cmd.strip ().split (' ')
+                exe = args[0]
+                exepath = pathcheck (exe)
+                if not exepath:
+                    break
+
+                # Main executable found.  But if it's 'gs',
+                # perhaps there is an IJS server we also need
+                # to check.
+                if os.path.basename (exepath) == 'gs':
+                    argn = len (args)
+                    argi = 1
+                    search = "-sIjsServer="
+                    while argi < argn:
+                        arg = args[argi]
+                        if arg.startswith (search):
+                            exe = arg[len (search):]
+                            exepath = pathcheck (exe)
+                            break
+
+                        argi += 1
+
+            if not exepath:
+                # Next pipe.
+                break
+
+    if exepath or not exe:
+        # Look for '*cupsFilter' lines in the PPD and check that
+        # the filters are installed.
+        (tmpfd, tmpfname) = tempfile.mkstemp ()
+        ppd.writeFd (tmpfd)
+        search = "*cupsFilter:"
+        for line in file (tmpfname).readlines ():
+            if line.startswith (search):
+                line = line[len (search):].strip ().strip ('"')
+                try:
+                    (mimetype, cost, exe) = line.split (' ')
+                except:
+                    continue
+
+                exepath = pathcheck (exe,
+                                     "/usr/lib/cups/filter:"
+                                     "/usr/lib64/cups/filter")
+
+    if exe and not exepath:
+        # We didn't find a necessary executable.  Complain.
+
+        # Strip out foomatic '%'-style place-holders.
+        p = exe.find ('%')
+        if p != -1:
+            exe = exe[:p]
+
+        pkgs = {
+            # Foomatic command line executables
+            'gs': 'ghostscript',
+            'perl': 'perl',
+            'foo2oak-wrapper': None,
+            'pnm2ppa': 'pnm2ppa',
+            'c2050': 'c2050',
+            'c2070': 'c2070',
+            'cjet': 'cjet',
+            'lm1100': 'lx',
+            'esc-m': 'min12xxw',
+            'min12xxw': 'min12xxw',
+            'pbm2l2030': 'pbm2l2030',
+            'pbm2l7k': 'pbm2l7k',
+            'pbm2lex': 'pbm2l7k',
+            # IJS servers (used by foomatic)
+            'hpijs': 'hpijs',
+            'ijsgutenprint.5.0': 'gutenprint',
+            # CUPS filters
+            'rastertogutenprint.5.0': 'gutenprint-cups',
+            'commandtoepson': 'gutenprint-cups',
+            'commandtocanon': 'gutenprint-cups',
+            }
+        try:
+            pkg = pkgs[exe]
+        except:
+            pkg = None
+
+        if pkg:
+            debugprint ("%s included in package %s" % (exe, pkg))
+            pkgs_to_install.append (pkg)
+        else:
+            exes_to_install.append (exe)
+
+    return (pkgs_to_install, exes_to_install)
 
 def main():
     c = cups.Connection()
